from .basic_normalizer import BasicNormalizer
from .detect import NswDetector
from .emoji_handler import EmojiHandler
from .normalize import ViSoLexNormalizer


def detect_nsw(input_str, model_repo=None, device='cpu'):
    """
    Detect Non-Standard Words (NSW) in the input text.
    
    Args:
        input_str (str): Input text to analyze
        model_repo (str, optional): HuggingFace model repository
        device (str): Device to run inference on ('cpu' or 'cuda')
    
    Returns:
        List of dictionaries containing NSW information
    """
    detector = NswDetector(model_repo=model_repo, device=device)
    nsw_spans = detector.detect_nsw(input_str)
    return nsw_spans

def normalize_sentence(input_str, detect_nsw=False, model_repo=None, device='cpu'):
    """
    Normalize a sentence and optionally detect NSW tokens.
    
    Args:
        input_str (str): Input text to normalize
        detect_nsw (bool): If True, also return NSW detection results
        model_repo (str, optional): HuggingFace model repository
        device (str): Device to run inference on ('cpu' or 'cuda')
    
    Returns:
        If detect_nsw=False: normalized text string
        If detect_nsw=True: tuple of (nsw_spans, normalized_text)
    """
    normalizer = ViSoLexNormalizer(model_repo=model_repo, device=device)
    return normalizer.normalize_sentence(input_str, detect_nsw)

__all__ = [
    'BasicNormalizer',
    'NswDetector', 
    'EmojiHandler',
    'ViSoLexNormalizer',
    'detect_nsw',
    'normalize_sentence'
]


import math
import numpy as np
import os
import gc

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from datasets import Dataset
from utils import gen_data_iter as gen_dataIter
from utils import get_config_attr

class RuleAttentionNetwork(nn.Module):
    def __init__(self, student_emb_dim, num_rules, num_labels, dense_dropout=0.3, max_rule_seq_length=2, device="cuda"):
        super(RuleAttentionNetwork, self).__init__()
        self.num_labels = num_labels
        self.device = device
        # Define Dropout Layer
        self.dropout = nn.Dropout(p=dense_dropout)

        # Dense Layer
        self.dense = nn.Linear(student_emb_dim, 128)
        # LayerNorm is more stable than BatchNorm for small effective batch sizes
        self.layer_norm = nn.LayerNorm(128)

        # Embedding Layers
        self.rule_embed = nn.Embedding(num_rules + 1, 128, padding_idx=0)
        self.rule_bias = nn.Embedding(num_rules + 1, 1, padding_idx=0)

        # Initialization
        nn.init.xavier_uniform_(self.rule_embed.weight) # Xavier uniform initializer
        nn.init.uniform_(self.rule_bias.weight)

        # Sigmoid Activation
        self.sigmoid = nn.Sigmoid()

    def forward(self, student_embeddings, rule_ids, rule_preds_onehot):
        # Apply Dropout
        x_hidden = self.dropout(student_embeddings)
        # Dense Layer + LayerNorm
        x_hidden = self.dense(x_hidden)
        x_hidden = self.layer_norm(x_hidden)
        x_hidden = F.relu(x_hidden)
        x_hidden = self.dropout(x_hidden)

        # Embedding Layers
        rule_embeddings = self.rule_embed(rule_ids)
        rule_biases = self.rule_bias(rule_ids)

        # Attention Scores
        att_scores = torch.matmul(x_hidden.unsqueeze(2), rule_embeddings.transpose(2, 3)).squeeze(2)
        att_scores = att_scores + rule_biases.squeeze(-1)
        att_sigmoid_proba = self.sigmoid(att_scores)
        
        outputs = torch.matmul(att_sigmoid_proba.float().unsqueeze(2), rule_preds_onehot.float()).squeeze(2)
        # Normalize Outputs
        outputs = normalize_with_random_rule(outputs, att_sigmoid_proba, rule_preds_onehot)
        outputs = l1_normalize(outputs, self.num_labels)

        return outputs, att_sigmoid_proba


class RAN:
    """
    Rule Attention Network
      * Input: text embedding x, array of rule predictions
      * Output: aggregate label
    """

    def __init__(self, args, num_rules, num_labels, logger=None, name='ran'):
        self.args = args
        self.name = name
        self.logger = logger
        self.manual_seed = get_config_attr(args, 'system.seed', get_config_attr(args, 'seed', 42))
        torch.manual_seed(self.manual_seed)
        training_mode = get_config_attr(args, 'model.training_mode', get_config_attr(args, 'training_mode', 'weakly_supervised'))
        default_logdir = './experiments/ran/{}'.format(training_mode)
        self.model_dir = get_config_attr(args, 'paths.logdir', get_config_attr(args, 'logdir', default_logdir))
        self.sup_batch_size = get_config_attr(args, 'training.train_batch_size', get_config_attr(args, 'train_batch_size', 16))
        self.eval_batch_size = get_config_attr(args, 'training.eval_batch_size', get_config_attr(args, 'eval_batch_size', 128))
        self.unsup_batch_size = get_config_attr(args, 'training.unsup_batch_size', get_config_attr(args, 'unsup_batch_size', 128))
        self.sup_epochs = get_config_attr(args, 'training.num_epochs', get_config_attr(args, 'num_epochs', 10))
        self.unsup_epochs = get_config_attr(args, 'training.num_unsup_epochs', get_config_attr(args, 'num_unsup_epochs', 5))
        self.num_labels = num_labels
        self.num_rules = num_rules
        self.device = get_config_attr(args, 'system.device', get_config_attr(args, 'device', 'cuda'))
        self.use_gpu = self.device=="cuda"

        # Using Student as an extra rule
        self.num_rules += 1
        self.student_rule_id = self.num_rules
        self.hard_student_rule = get_config_attr(args, 'model.hard_student_rule', get_config_attr(args, 'hard_student_rule', False))
        self.trained = False
        self.xdim = None
        self.ignore_student = False

    def postprocess_rule_preds(self, rule_pred, student_pred=None):
        """Torch-only preprocessing on device.
        Inputs:
          - rule_pred: LongTensor [B, T, R_rules], entries in [0..num_labels-1] or -1 for no rule
          - student_pred: None or FloatTensor [B, T, num_labels] (soft) or [B, T] (hard ids)
        Returns:
          - rule_mask: FloatTensor [B, T, S] where S = R_rules (+1 if student included)
          - fired_rule_ids: LongTensor [B, T, S] with 0 as padding, student id where applicable
          - one_hot_rule_pred: FloatTensor [B, T, S, num_labels]
        """
        device = rule_pred.device
        B, T, Rrules = rule_pred.shape

        # Mask of fired rules
        rule_mask = (rule_pred != -1).to(torch.float32)  # [B, T, Rrules]

        # Fired rule ids: 1..Rrules where mask True, else 0 (padding)
        rule_indices = (torch.arange(Rrules, device=device) + 1).view(1, 1, Rrules)
        fired_rule_ids = torch.where(rule_mask.bool(), rule_indices, torch.zeros_like(rule_indices))  # [B, T, Rrules]

        # One-hot predictions per rule (zeros where -1)
        clamped = torch.clamp(rule_pred, min=0)
        oh = F.one_hot(clamped, num_classes=self.num_labels).to(torch.float32)  # [B, T, Rrules, C]
        oh = oh * rule_mask.unsqueeze(-1)

        # Optionally prepend student rule
        if student_pred is not None:
            # Align time dimension T between student and rules by cropping to min length
            T_student = student_pred.size(1) if student_pred.dim() >= 2 else T
            T_common = min(T, T_student)
            if T_common != T:
                rule_mask = rule_mask[:, :T_common, :]
                fired_rule_ids = fired_rule_ids[:, :T_common, :]
                oh = oh[:, :T_common, :, :]
                T = T_common
            if student_pred.dim() == 3:  # soft probs
                student_pred = student_pred[:, :T, :]
                student_one_hot = student_pred.to(torch.float32)
            else:
                student_pred = student_pred[:, :T]
                # hard ids -> one-hot
                student_one_hot = F.one_hot(student_pred.to(torch.long), num_classes=self.num_labels).to(torch.float32)
            if student_pred.dim() == 3:  # soft probs
                student_one_hot = student_pred.to(torch.float32)
            else:
                # hard ids -> one-hot
                student_one_hot = F.one_hot(student_pred.to(torch.long), num_classes=self.num_labels).to(torch.float32)
            student_one_hot = student_one_hot.unsqueeze(2)  # [B, T, 1, C]

            student_mask = torch.ones((B, T, 1), device=device, dtype=torch.float32)
            student_ids = torch.full((B, T, 1), fill_value=self.student_rule_id if not self.ignore_student else 0, device=device, dtype=torch.long)

            one_hot_rule_pred = torch.cat([student_one_hot, oh], dim=2)          # [B, T, 1+Rrules, C]
            rule_mask = torch.cat([student_mask, rule_mask], dim=2)              # [B, T, 1+Rrules]
            fired_rule_ids = torch.cat([student_ids, fired_rule_ids], dim=2)     # [B, T, 1+Rrules]
        else:
            one_hot_rule_pred = oh

        return rule_mask, fired_rule_ids, one_hot_rule_pred

    def init_model(self):
        if self.xdim is None:
            raise ValueError("xdim must be set before initializing the model. Call _setup_teacher_dimensions() first.")
        self.model = RuleAttentionNetwork(self.xdim,
                                          num_rules=self.num_rules,
                                          num_labels=self.num_labels,
                                          max_rule_seq_length=self.num_rules,
                                          device=self.device)
        if self.use_gpu:
            self.model = self.model.to(self.device)

    def _ensure_model_initialized(self):
        """Ensure the model is initialized before use."""
        if not hasattr(self, 'model') or self.model is None:
            self.init_model()

    def epoch_run(self, data, mode, loss_fn, optimizer=None, scheduler=None, gradient_accumulation_steps=4, scaler=None):
        # mode in ["sup_train", "unsup_train", "dev"]
        total_loss = 0
        num_batch = len(data['id'])
        
        # Initialize gradient accumulation
        if mode in ['sup_train', 'unsup_train']:
            optimizer.zero_grad()
        
        for i in range(num_batch):
            # Process one pre-batched chunk (already batched by upstream pipeline)
            with torch.amp.autocast(device_type='cuda' if self.use_gpu else 'cpu', enabled=True):
                x = torch.as_tensor(data['features'][i], dtype=torch.float32, device=self.device)
                student_pred_np = data['proba'][i]
                rule_pred_np = data['weak_labels'][i]

                # Move rule/student preds to tensors on device
                rule_pred = torch.as_tensor(rule_pred_np, dtype=torch.long, device=self.device)
                student_pred = None
                if student_pred_np is not None:
                    if isinstance(student_pred_np, np.ndarray):
                        student_pred = torch.as_tensor(student_pred_np, dtype=torch.float32, device=self.device)
                    else:
                        # assume already a tensor-like list -> convert
                        student_pred = torch.as_tensor(student_pred_np, dtype=torch.float32, device=self.device)

                rule_mask, fired_rule_ids, rule_one_hot = self.postprocess_rule_preds(rule_pred, student_pred)
                
                if mode != 'unsup_train':
                    y = torch.as_tensor(data['output_ids'][i], dtype=torch.long, device=self.device)
                
                if mode in ['sup_train', 'unsup_train']:
                    outputs, _ = self.model(x, fired_rule_ids, rule_one_hot)
                    loss = loss_fn(outputs) if mode=='unsup_train' else loss_fn(outputs.view(-1, self.num_labels), y.view(-1))
                    
                    # Scale loss for gradient accumulation
                    loss = loss / gradient_accumulation_steps
                    total_loss += loss.detach() * gradient_accumulation_steps
                    
                    if scaler is not None:
                        scaler.scale(loss).backward()
                    else:
                        loss.backward()
                    
                    # Update weights every gradient_accumulation_steps
                    if (i + 1) % gradient_accumulation_steps == 0:
                        if scaler is not None:
                            scaler.step(optimizer)
                            scaler.update()
                        else:
                            optimizer.step()
                        if scheduler:
                            scheduler.step()
                        optimizer.zero_grad(set_to_none=True)
                        
                else:
                    outputs_dev, _ = self.model(x, fired_rule_ids, rule_one_hot)
                    dev_loss = loss_fn(outputs_dev.view(-1, self.num_labels), y.view(-1))
                    total_loss += dev_loss.detach()
            
            # Clear intermediate tensors to free memory
            del x, fired_rule_ids, rule_pred, rule_mask, rule_one_hot
            if student_pred is not None:
                del student_pred
            if mode != 'unsup_train':
                del y
            if mode in ['sup_train', 'unsup_train']:
                del outputs, loss
            else:
                del outputs_dev, dev_loss
            
            # Periodic memory cleanup every 20 batches to prevent accumulation
            if (i + 1) % 20 == 0:
                if self.use_gpu:
                    torch.cuda.empty_cache()
                import gc
                gc.collect()
        if mode == 'unsup_train':
            scheduler.step()
        if mode == 'sup_train':
            scheduler.step(total_loss/num_batch)
        
        return total_loss/num_batch

    def train(self, train_data, dev_data=None, unsup_data=None, gradient_accumulation_steps=4):
        assert unsup_data is not None, "For SSL RAN you need to also provide unlabeled data... "

        self._ensure_model_initialized()
        if not self.trained:
            pass  # Model already initialized above
        
        self.logger.info("\n\n\t\t*** Training RAN ***")
        self.logger.info(f"Using gradient accumulation with {gradient_accumulation_steps} steps")

        # Log initial memory usage
        self._log_memory_usage("Before training")

        loss_fn = MinEntropyLoss()
        optimizer = optim.Adam(self.model.parameters(), lr=1e-3, weight_decay=1e-5)  # Add weight decay
        scaler = torch.amp.GradScaler('cuda', enabled=self.use_gpu)
        scheduler = create_learning_rate_scheduler(optimizer,
                                                   max_learn_rate=1e-2,
                                                   end_learn_rate=1e-5,
                                                   warmup_epoch_count=2,
                                                   total_epoch_count=self.sup_epochs)

        # Training loop for unsupervised data
        self.model.train()
        unsup_losses = []
        for epoch in range(self.sup_epochs):
            unsup_loss = self.epoch_run(unsup_data, mode="unsup_train", loss_fn=loss_fn, optimizer=optimizer, scheduler=scheduler, gradient_accumulation_steps=gradient_accumulation_steps, scaler=scaler)
            unsup_losses.append(unsup_loss.item())
            self.logger.info("Unsupervised trainning: EPOCH {}/{} - LOSS: {}".format(epoch+1, self.sup_epochs, unsup_loss))
            
            # Memory cleanup after each epoch
            self._cleanup_epoch_memory()
            
            # Log memory usage every 3 epochs
            if (epoch + 1) % 3 == 0:
                self._log_memory_usage(f"After unsup epoch {epoch+1}")

        # Reinitialize loss function and optimizer for supervised training
        loss_fn = nn.CrossEntropyLoss(ignore_index=self.num_labels)
        optimizer = optim.Adam(self.model.parameters(), lr=1e-3, weight_decay=1e-5)  # Add weight decay
        scaler = torch.amp.GradScaler('cuda', enabled=self.use_gpu)
        scheduler = create_learning_rate_scheduler(optimizer,
                                                   max_learn_rate=1e-2,
                                                   end_learn_rate=1e-5,
                                                   warmup_epoch_count=2,
                                                   total_epoch_count=self.sup_epochs)
        
        # Training loop for supervised training
        self.trained = True
        train_losses = []
        dev_losses = []
        best_val_loss = np.inf
        for epoch in range(self.sup_epochs):
            self.model.train()
            train_loss = self.epoch_run(train_data, mode="sup_train", loss_fn=loss_fn, optimizer=optimizer, scheduler=scheduler, gradient_accumulation_steps=gradient_accumulation_steps, scaler=scaler)
            train_losses.append(train_loss.item())
            self.model.eval()
            with torch.no_grad():
                dev_loss = self.epoch_run(dev_data, mode="dev", loss_fn=loss_fn)
                dev_losses.append(dev_loss.item())
                if dev_loss < best_val_loss:
                    best_val_loss = dev_loss
                    best_model = self.model
            
            # Memory cleanup after each epoch
            self._cleanup_epoch_memory()
            
            # Log memory usage every 3 epochs
            if (epoch + 1) % 3 == 0:
                self._log_memory_usage(f"After sup epoch {epoch+1}")
                
            self.logger.info("Supervised trainning: EPOCH {}/{} - TRAIN LOSS: {} - VAL LOSS: {} - BEST VAL LOSS: {}".format(epoch+1, self.sup_epochs, train_loss, dev_loss, best_val_loss))
        self.model = best_model
        
        # Final memory cleanup
        self._cleanup_epoch_memory()
        self._log_memory_usage("After training")
        
        return {
            'unsup_loss': unsup_losses,
            'sup_loss': train_losses,
            'dev_loss': dev_losses,
            'best_dev_loss': best_val_loss.item(),
        }
    
    def _cleanup_epoch_memory(self):
        """Clean up memory after each epoch."""
        import gc
        
        # Clear GPU cache
        if self.use_gpu:
            torch.cuda.empty_cache()
        
        # Force garbage collection
        gc.collect()
        
        # Clear any cached computations
        if hasattr(self.model, 'cache'):
            delattr(self.model, 'cache')
    
    def _process_data_in_chunks(self, data, chunk_size=100):
        """Process large datasets in chunks to reduce memory usage."""
        total_size = len(data['id'])
        for start_idx in range(0, total_size, chunk_size):
            end_idx = min(start_idx + chunk_size, total_size)
            chunk = {key: data[key][start_idx:end_idx] for key in data.keys()}
            yield chunk

    def predict_ran(self, dataset, batch_size=128, inference_mode=False):
        self._ensure_model_initialized()
        
        y_preds = []
        soft_probas = []

        label = False
        if 'output_ids' in dataset:
            label=True
        num_batch = len(dataset['id'])

        self.model.eval()
        with torch.no_grad():
            for i in range(num_batch):
                # Use mixed precision for inference to save memory
                with torch.amp.autocast(device_type='cuda' if self.use_gpu else 'cpu', enabled=True):
                    x_batch = torch.as_tensor(dataset['features'][i], dtype=torch.float32, device=self.device)
                    rule_pred_batch = torch.as_tensor(dataset['weak_labels'][i], dtype=torch.long, device=self.device)
                    student_pred_np = dataset['proba'][i]

                    # Align all time dimensions to common minimum to avoid shape mismatches
                    T_feat = x_batch.shape[0]
                    T_rule = rule_pred_batch.shape[0]
                    T_student = student_pred_np.shape[0] if student_pred_np is not None else T_rule
                    T_common = min(T_feat, T_rule, T_student)
                    if T_feat != T_common:
                        x_batch = x_batch[:T_common]
                    if T_rule != T_common:
                        rule_pred_batch = rule_pred_batch[:T_common]
                    if student_pred_np is not None and T_student != T_common:
                        student_pred_np = student_pred_np[:T_common]

                    if student_pred_np is None:
                        random_pred = (rule_pred_batch != -1).sum(dim=-1) == 0  # [T_common]
                    else:
                        # Track to-mask vector per token (length T_common)
                        random_pred = torch.zeros((rule_pred_batch.shape[0],), dtype=torch.bool, device=self.device)
                    rule_mask, fired_rule_ids, rule_pred_one_hot = self.postprocess_rule_preds(rule_pred_batch, torch.as_tensor(student_pred_np, dtype=torch.float32, device=self.device) if student_pred_np is not None else None)
                
                    y_pred, att_score = self.model(x_batch, fired_rule_ids, rule_pred_one_hot)

                    preds = torch.argmax(y_pred, dim=-1, keepdim=False)  # [T_common]
                    max_proba = torch.max(y_pred, dim=-1)[0]
                    if max_proba.dim() > 1:
                        max_proba = max_proba.squeeze()
                    confidence_thres = 0.5
                    ignore_pred = (max_proba < confidence_thres)
                    # Ensure 1-D bool mask
                    if ignore_pred.dim() > 1:
                        ignore_pred = ignore_pred.reshape(-1)
                    ignore_pred = ignore_pred.to(torch.bool)
                    # Align mask and vectors to same length
                    L = min(random_pred.shape[0], ignore_pred.shape[0], preds.shape[0])
                    if random_pred.shape[0] != L:
                        random_pred = random_pred[:L]
                    if ignore_pred.shape[0] != L:
                        ignore_pred = ignore_pred[:L]
                    if preds.shape[0] != L:
                        preds = preds[:L]
                    # Use index list to avoid dimensionality issues in boolean indexing
                    idx_to_ignore = torch.nonzero(ignore_pred, as_tuple=False).view(-1)
                    if idx_to_ignore.numel() > 0:
                        random_pred[idx_to_ignore] = True
                    preds[random_pred] = -1
                if not inference_mode:
                    soft_proba = y_pred

                # Convert to CPU and store immediately to reduce GPU memory
                y_preds.append(preds.detach().cpu().numpy())
                if not inference_mode:
                    soft_probas.append(soft_proba.detach().cpu().numpy())
                
                # Clear intermediate tensors to free memory
                del x_batch, rule_pred_batch, random_pred
                del fired_rule_ids, rule_pred_one_hot, y_pred, att_score, preds, max_proba
                if not inference_mode:
                    del soft_proba
                
                # Periodic memory cleanup every 10 batches
                if (i + 1) % 10 == 0:
                    self._cleanup_epoch_memory()

        if inference_mode:
            return {
                'id': dataset['id'],
                'input_ids': dataset['input_ids'],
                'align_index': dataset['align_index'],
                'preds': y_preds,
            }

        return {
            'id': dataset['id'],
            'input_ids': dataset['input_ids'],
            'output_ids': dataset['output_ids'] if label else None,
            'align_index': dataset['align_index'],
            'preds': y_preds,
            'proba': soft_probas,
        }

    def _log_memory_usage(self, stage=""):
        """Log current memory usage for debugging."""
        if self.use_gpu and torch.cuda.is_available():
            allocated = torch.cuda.memory_allocated() / 1024**3  # GB
            reserved = torch.cuda.memory_reserved() / 1024**3    # GB
            self.logger.info(f"Memory usage {stage}: Allocated={allocated:.2f}GB, Reserved={reserved:.2f}GB")
        else:
            import psutil
            process = psutil.Process(os.getpid())
            memory_mb = process.memory_info().rss / 1024**2
            self.logger.info(f"Memory usage {stage}: {memory_mb:.2f}MB")

    def load(self, savefile):
        self._ensure_model_initialized()
        self.logger.info("loading rule attention network from {}".format(savefile))
        self.model.load_state_dict(torch.load(savefile))

    def save(self, savefile):
        self._ensure_model_initialized()
        self.logger.info("Saving rule attention network at {}".format(savefile))
        torch.save(self.model.state_dict(), savefile)
        return


def MinEntropyLoss():
    def loss(y_prob):
        per_example_loss = -y_prob * torch.log(y_prob)
        return torch.mean(per_example_loss)
    return loss

def create_learning_rate_scheduler(optimizer,
                                   max_learn_rate=1e-2,
                                   end_learn_rate=1e-5,
                                   warmup_epoch_count=3,
                                   total_epoch_count=10):
    def lr_scheduler(epoch):
        if epoch < warmup_epoch_count:
            res = (max_learn_rate / warmup_epoch_count) * (epoch + 1)
        else:
            # Avoid division by zero
            decay_epochs = total_epoch_count - warmup_epoch_count
            if decay_epochs <= 0:
                # If no decay epochs, just return the max learning rate
                res = max_learn_rate
            else:
                res = max_learn_rate * math.exp(
                    math.log(end_learn_rate / max_learn_rate) * (epoch - warmup_epoch_count + 1) / decay_epochs)
        return float(res)

    learning_rate_scheduler = optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lr_scheduler)

    return learning_rate_scheduler

def l1_normalize(x, num_labels):
    x = x + 1e-05  # Avoid stability issues
    l1_norm = torch.sum(x, dim=-1, keepdim=True).detach()
    l1_norm = torch.repeat_interleave(l1_norm, repeats=num_labels, dim=-1)  # Equivalent to tf.keras.backend.repeat_elements()

    return x / l1_norm

def normalize_with_random_rule(output, att_sigmoid_proba, rule_preds_onehot):
    device=output.device
    num_labels = rule_preds_onehot.shape[-1]
    sum_prob = torch.sum(rule_preds_onehot, dim=-1).detach()
    rule_mask = (sum_prob > 0).float()
    num_rules = torch.sum(sum_prob, dim=-1).float()
    masked_att_proba = att_sigmoid_proba * rule_mask
    sum_masked_att_proba = torch.sum(masked_att_proba, dim=-1)
    uniform_rule_att_proba = num_rules - sum_masked_att_proba
    uniform_vec = torch.ones((uniform_rule_att_proba.shape[0], uniform_rule_att_proba.shape[1], num_labels)) / num_labels
    uniform_vec = uniform_vec.to(device)
    uniform_pred = torch.repeat_interleave(uniform_rule_att_proba.unsqueeze(-1), repeats=num_labels, dim=-1) * uniform_vec
    output_with_uniform_rule = output + uniform_pred
    return output_with_uniform_rule



"""
Configuration management for ViSoNorm project.
Centralizes all configuration parameters and constants.
"""
import os
from dataclasses import dataclass
from typing import Dict, List, Optional
from enum import Enum

# Project path
PROJECT_PATH = os.path.dirname(os.path.realpath(__file__))


class TrainingMode(Enum):
    ONLY_STUDENT = "only_student"
    SELF_TRAINING = "self_training"
    WEAKLY_SUPERVISED = "weakly_supervised"


class BaseModel(Enum):
    PHOBERT = "phobert"
    VISOBERT = "visobert"
    BARTPHO = "bartpho"
    VIT5 = "vit5"


class FineTuningStrategy(Enum):
    STANDARD = "standard"
    FLEXIBLE_LR = "flexible_lr"


@dataclass
class ModelConfig:
    """Configuration for model parameters."""
    base_model: BaseModel = BaseModel.VISOBERT
    training_mode: TrainingMode = TrainingMode.WEAKLY_SUPERVISED
    inference_model: str = "student"
    
    # Model architecture
    num_rules: int = 2
    append_n_mask: bool = False
    nsw_detector: bool = False  # NSW (Non-Standard Word) detection
    soft_labels: bool = False
    loss_weights: bool = False
    hard_student_rule: bool = False
    
    # HuggingFace pretrained model name (unified for all backbones).
    # If None or empty, a sensible default will be chosen based on `base_model`:
    #   phobert -> "vinai/phobert-base"
    #   visobert -> "uitnlp/visobert"
    #   bartpho -> "vinai/bartpho-syllable"
    #   vit5    -> "VietAI/vit5-base"
    pretrained_model_name: Optional[str] = None
    
    # Multi-task training weights
    nsw_loss_weight: float = 1.0
    n_mask_loss_weight: float = 1.0
    norm_loss_weight: float = 1.0
    
    # Fine-tuning
    fine_tuning_strategy: FineTuningStrategy = FineTuningStrategy.FLEXIBLE_LR
    learning_rate: float = 1e-3
    
    # Optimizer settings
    optimizer: str = "adamw"  # "adam", "adamw"
    weight_decay: float = 0.01
    adam_betas: tuple = (0.9, 0.999)
    adam_eps: float = 1e-8
    
    # Learning rate scheduling
    use_scheduler: bool = True
    scheduler_type: str = "cosine"  # "cosine", "linear", "plateau", "none"
    warmup_epochs: int = 2
    min_lr_ratio: float = 0.1  # min_lr = learning_rate * min_lr_ratio
    
    # Gradient clipping
    max_grad_norm: float = 1.0


@dataclass
class TrainingConfig:
    """Configuration for training parameters."""
    num_epochs: int = 10
    num_unsup_epochs: int = 5
    num_iter: int = 10
    
    # Batch sizes
    train_batch_size: int = 16
    eval_batch_size: int = 128
    unsup_batch_size: int = 128
    
    # Data processing
    sample_size: int = 8096
    remove_accents: bool = False
    rm_accent_ratio: float = 0.5
    lower_case: bool = False
    
    # Evaluation
    metric: str = "f1_score"


@dataclass
class PathConfig:
    """Configuration for file paths."""
    project_path: str = os.path.dirname(os.path.realpath(__file__))
    datapath: str = os.path.join(project_path, "data")
    experiment_folder: str = os.path.join(project_path, "experiments")
    logdir: str = "./experiments/visobert"
    
    @property
    def pretrained_phobert(self) -> str:
        return os.path.join(self.project_path, "model/pretrained/phobert")
    
    @property
    def pretrained_visobert(self) -> str:
        return os.path.join(self.project_path, "model/pretrained/visobert")
    
    @property
    def sp_model(self) -> str:
        return os.path.join(self.pretrained_visobert, "sentencepiece.bpe.model")


@dataclass
class SystemConfig:
    """Configuration for system parameters."""
    seed: int = 42
    debug: bool = False
    device: Optional[str] = None
    n_gpu: int = 0


@dataclass
class Config:
    """Main configuration class combining all configs."""
    model: ModelConfig = ModelConfig()
    training: TrainingConfig = TrainingConfig()
    paths: PathConfig = PathConfig()
    system: SystemConfig = SystemConfig()
    
    def __post_init__(self):
        """Set up derived configurations after initialization."""
        self.system.logdir = os.path.join(
            self.paths.experiment_folder, 
            self.model.base_model.value, 
            self.model.training_mode.value
        )


# Tokenizer constants
MASK_TOKEN = '<mask>'
BOS_TOKEN = '<s>'
EOS_TOKEN = '</s>'
PAD_TOKEN = '<pad>'
NULL_STR = '<space>'
UNK_TOKEN = '<unk>'
NUM_LABELS_N_MASKS = 5
NULL_STR_TO_SHOW = '_'

SPECIAL_TOKEN_LS = [NULL_STR, MASK_TOKEN, BOS_TOKEN, EOS_TOKEN, UNK_TOKEN]

PRETRAINED_TOKENIZER_MAP = {
    'phobert': 'vinai/phobert-base',
    'visobert': 'uitnlp/visobert',
    'bartpho': 'vinai/bartpho-syllable',
    'vit5': 'VietAI/vit5-base',
}

# Pretrained model paths
PRETRAINED_VISOBERT = os.path.join(PROJECT_PATH, "model/pretrained/visobert")
PRETRAINED_PHOBERT = os.path.join(PROJECT_PATH, "model/pretrained/phobert")

# Accent removal dictionary
RM_ACCENTS_DICT = str.maketrans(
    "ÁÀẢÃẠĂẮẰẲẴẶÂẤẦẨẪẬĐÈÉẺẼẸÊẾỀỂỄỆÍÌỈĨỊÓÒỎÕỌÔỐỒỔỖỘƠỚỜỞỠỢÚÙỦŨỤƯỨỪỬỮỰÝỲỶỸỴáàảãạăắằẳẵặâấầẩẫậđèéẻẽẹêếềểễệíìỉĩịóòỏõọôốồổỗộơớờởỡợúùủũụưứừửữựýỳỷỹỵ",
    "A"*17 + "D" + "E"*11 + "I"*5 + "O"*17 + "U"*11 + "Y"*5 + "a"*17 + "d" + "e"*11 + "i"*5 + "o"*17 + "u"*11 + "y"*5
)

# Available fine-tuning strategies
AVAILABLE_FINE_TUNING_STRATEGY = ["standard", "flexible_lr"]

# Additional constants for backward compatibility
DATASET_DIR = os.path.join(PROJECT_PATH, "data")
SP_MODEL = os.path.join(PRETRAINED_VISOBERT, "sentencepiece.bpe.model")
NUM_RULES = 2
EXPERIMENT_DIR = os.path.join(PROJECT_PATH, "experiments")
UNLABELED_SAMPLE_SIZE = 8096

# Samples per task for reporting
SAMPLES_PER_TASK_TO_REPORT = {
    "normalize": ["all", "NEED_NORM", "NORMED", "PRED_NEED_NORM", "PRED_NORMED"],
    "n_masks_pred": ["all", "n_masks_1", "n_masks_2", "n_masks_3", "n_masks_4", "n_masks_5"]
}



import os
from torch.utils.data import Dataset
import numpy as np
import pandas as pd
import torch
from copy import deepcopy
from sklearn.utils import shuffle
from ast import literal_eval
from itertools import chain
from utils import add_special_tokens as add_special_token, run_strip_accents
from utils import get_config_attr
from config import MASK_TOKEN, NULL_STR


class DataHandler:
    # This module is responsible for feeding the data to teacher/student
    # If teacher is applied, then student gets the teacher-labeled data instead of ground-truth labels
    def __init__(self, args, tokenizer, logger=None):
        self.args = args
        self.logger = logger
        self.tokenizer = tokenizer
        self.datasets = {}
        self.seed = get_config_attr(args, 'system.seed', get_config_attr(args, 'seed', 42))
        np.random.seed(self.seed)

    def load_dataset(self, method='train'):
        dataset = WSDataset(self.args, method=method, tokenizer=self.tokenizer, logger=self.logger)
        self.datasets[method] = dataset
        return dataset

    def create_pseudodataset(self, wsdataset):
        dataset = PseudoDataset(self.args, wsdataset, self.logger)
        return dataset


class WSDataset(Dataset):
    # WSDataset: Dataset for Weak Supervision.
    def __init__(self, args, method, tokenizer, logger=None):
        super(WSDataset, self).__init__()
        self.args = args
        self.seed = get_config_attr(args, 'system.seed', get_config_attr(args, 'seed', 42))
        self.lower_case = get_config_attr(args, 'training.lower_case', get_config_attr(args, 'lower_case', False))
        self.remove_accents = get_config_attr(args, 'training.remove_accents', get_config_attr(args, 'remove_accents', False))
        self.rm_accent_ratio = get_config_attr(args, 'training.rm_accent_ratio', get_config_attr(args, 'rm_accent_ratio', 0.5))
        
        datapath = get_config_attr(args, 'paths.datapath', get_config_attr(args, 'datapath', './data'))
        self.datapath = os.path.join(datapath, "{}.csv".format(method))
        
        self.method = method
        self.tokenizer = tokenizer
        self.logger = logger
        self.data = {}
        self.no_accent_data = {}
        self.load_dataset()
        self.num_labels = len(self.tokenizer)
    
    def strip_accents(self, sent_list):
        new_sent_list = []
        for sent in sent_list:
            word_list = []
            for word in sent:
                new_word = run_strip_accents(word, self.rm_accent_ratio)
                word_list.append(new_word)
            new_sent_list.append(word_list)
        return new_sent_list
    
    def preprocess(self, data, strip_accents=False):
        tokenized_src_ls = []
        tokenized_tgt_ls = [] if self.method != "unlabeled" else None
        aligned_idx_ls = []
        weak_label_ls = []
        sent_len_ls = []

        ids = data['id'].tolist()
        inputs = self.strip_accents(data['input'].tolist()) if strip_accents else data['input'].tolist()
        outputs = data['output'].tolist() if self.method != "unlabeled" else None
        weak_rules = [col for col in data.columns if col.startswith("rule")]
        weak_labels = data[weak_rules].values
        num_sents = len(data)
        num_rules = weak_labels.shape[1]

        for i in range(num_sents):
            tokenized_src = []
            tokenized_tgt = []
            aligned_idx = []
            tokenized_rule_preds = []

            input_seq = inputs[i]
            output_seq = outputs[i] if self.method != "unlabeled" else None
            rule_preds = weak_labels[i]
            
            if self.lower_case:
                input_seq = [token.lower() for token in input_seq]
                output_seq = [token.lower() for token in output_seq] if self.method != "unlabeled" else None
                for j, preds in enumerate(rule_preds):
                    if preds is None:
                        continue
                    
                    # Check for None values and non-string values in preds before calling .lower()
                    filtered_preds = []
                    for k, token in enumerate(preds):
                        if token is None:
                            filtered_preds.append("")  # Replace None with empty string
                        elif not isinstance(token, str):
                            filtered_preds.append(str(token).lower())  # Convert to string first, then lowercase
                        else:
                            filtered_preds.append(token.lower())
                    rule_preds[j] = filtered_preds
            aligned = 0
            for idx, source_token in enumerate(input_seq):
                target_token = output_seq[idx] if self.method != "unlabeled" else None
                rule_pred = [pred[idx] for pred in rule_preds]
                tokenized_source = self.tokenizer.tokenize(source_token)
                tokenized_target = self.tokenizer.tokenize(target_token) if self.method != "unlabeled" else None
 
                if self.method != "unlabeled":
                    len_diff = len(tokenized_source) - len(tokenized_target)
                    if len_diff < 0:
                        tokenized_source.extend([MASK_TOKEN]*abs(len_diff))
                    elif len_diff > 0:
                        tokenized_target.extend([NULL_STR]*len_diff)
                
                tokenized_rule_pred = [self.tokenizer.tokenize(token) for token in rule_pred]
                for j, pred in enumerate(tokenized_rule_pred):
                    if len(pred) < len(tokenized_source):
                        tokenized_rule_pred[j].extend([NULL_STR]*(len(tokenized_source)-len(pred)))
                    elif len(pred) > len(tokenized_source):
                        # print("WARNING: INDEX {} - LENGTH OF WEAK LABELS {} IS LONGER THAN SOURCE TOKEN {}".format(ids[i], pred, tokenized_source))
                        tokenized_rule_pred[j] = pred[:len(tokenized_source)]

                tokenized_src.extend(tokenized_source)
                if self.method != "unlabeled":
                    tokenized_tgt.extend(tokenized_target)
                aligned_idx.extend([aligned]*len(tokenized_source))
                aligned += 1

                if not tokenized_rule_preds:
                    tokenized_rule_preds = tokenized_rule_pred
                else:
                    for j in range(num_rules):
                        tokenized_rule_preds[j].extend(tokenized_rule_pred[j])

            add_special_token(tokenized_src)
            if self.method != "unlabeled":
                add_special_token(tokenized_tgt)
            for j in range(len(tokenized_rule_preds)):
                add_special_token(tokenized_rule_preds[j])

            sent_len = len(tokenized_src)
            
            tokenized_src_ls.append(tokenized_src)
            if self.method != "unlabeled":
                tokenized_tgt_ls.append(tokenized_tgt)
            aligned_idx_ls.append(aligned_idx)
            weak_label_ls.append(tokenized_rule_preds)
            sent_len_ls.append(sent_len)

        # Converts tokens to ids
        input_ids = [self.tokenizer.convert_tokens_to_ids(sent) for sent in tokenized_src_ls]
        output_ids = [self.tokenizer.convert_tokens_to_ids(sent) for sent in tokenized_tgt_ls] if self.method != "unlabeled" else None
        for i, weak_labels in enumerate(weak_label_ls):
            weak_label_ls[i] = [self.tokenizer.convert_tokens_to_ids(sent) for sent in weak_labels]

        reshaped_weak_label = []
        for i, sent in enumerate(weak_label_ls):
            num_rules = len(sent)
            num_words = len(sent[0])
            new_sent = []
            for j in range(num_words):
                new_sent.append([rule[j] for rule in sent])
            reshaped_weak_label.append(new_sent)
        reshaped_weak_label
          
        if self.method == "unlabeled":
            preprocessed_dataset = {
                'id': ids,
                'input': inputs,
                'input_ids': input_ids,
                'align_index': aligned_idx_ls,
                'weak_labels': reshaped_weak_label,
                'sent_len': sent_len_ls,
            }
        else:
          preprocessed_dataset = {
              'id': ids,
              'input': inputs,
              'output': outputs,
              'input_ids': input_ids,
              'output_ids': output_ids,
              'align_index': aligned_idx_ls,
              'weak_labels': reshaped_weak_label,
              'sent_len': sent_len_ls,
          }
        return preprocessed_dataset

    def load_dataset(self):
        if self.method == "unlabeled":
            converters = {'input': literal_eval, 'regex_rule': literal_eval, 'dict_rule': literal_eval, 'gpt4omini_rule': literal_eval, 'gemini_rule': literal_eval}
        else:
            converters = {'input': literal_eval, 'output': literal_eval, 'regex_rule': literal_eval, 'dict_rule': literal_eval, 'gpt4omini_rule': literal_eval, 'gemini_rule': literal_eval}
        
        data = pd.read_csv(self.datapath, converters=converters)
        data.rename(columns={'regex_rule': 'rule_01', 'dict_rule': 'rule_02', 'gpt4omini_rule': 'rule_03', 'gemini_rule': 'rule_04'}, inplace=True)
        data = data.dropna(ignore_index=True)

        self.logger.info("Pre-processing {} data for student...".format(self.method))
        self.data = self.preprocess(data) # dictionary: column - list of values (each_sentence)
        if self.remove_accents and self.method != 'unlabeled':
            self.logger.info("Removing accents of {} data for student...".format(self.method))
            self.no_accent_data = self.preprocess(data, strip_accents=True)
            new_dict = {}
            for key, values in self.no_accent_data.items():
                new_dict[key] = values + self.data[key]
            self.no_accent_data = new_dict

    def __len__(self):
        return len(self.data['id'])

    def __getitem__(self, item):
        ret = {
            'id': self.data['id'][item],
            'input': self.data['input'][item],
            'output': self.data['output'][item] if 'output' in self.data else None,
            'input_ids': self.data['input_ids'][item],
            'output_ids': self.data['output_ids'][item] if 'output' in self.data else None,
            'align_index': self.data['align_index'][item],
            'weak_labels': self.data['weak_labels'][item],
            'sent_len': self.data['sent_len'][item],
            }
        return ret


class PseudoDataset(Dataset):
    # PseudoDataset: a Dataset class that provides extra functionalities for teacher-student training.
    def __init__(self, args, wsdataset, logger=None):
        super(PseudoDataset, self).__init__()
        self.args = args
        self.seed = get_config_attr(args, 'system.seed', get_config_attr(args, 'seed', 42))
        self.method = wsdataset.method
        self.logger = logger
        self.num_labels = wsdataset.num_labels
        self.logger.info("copying data from {} dataset".format(wsdataset.method))
        self.original_data = deepcopy(wsdataset.data)
        self.data = deepcopy(self.original_data)
        self.no_accent_data = deepcopy(wsdataset.no_accent_data)
        self.student_data = {}
        self.teacher_data = {}
        self.logger.info("done")

    def keep(self, keep_indices, type='teacher'):
        self.logger.info("Creating Pseudo Dataset with {} items...".format(len(list(chain.from_iterable(keep_indices)))))
        new_dict = {}
        data = self.teacher_data if type=='teacher' else self.student_data
        for key, values in data.items():
            keep_sents = []
            for i, indices in enumerate(keep_indices):
                if len(indices)==0:
                    continue
                if key in ['id', 'align_index']:
                    keep_sent = [values[i][idx] for idx in indices]
                else:
                    keep_sent = values[i][np.array(indices)]
                keep_sents.append(keep_sent)
            new_dict[key] = keep_sents
        if type=='teacher':
            self.teacher_data = new_dict
        else:
            self.student_data = new_dict

    def downsample(self, sample_size):
        N = len(self.original_data['input'])
        if sample_size > N:
            self.logger.info("[WARNING] sample size = {} > {}".format(sample_size, N))
            sample_size = N
        self.logger.info("Downsampling {} data".format(sample_size))
        self.data = {}
        keep_indices = np.random.choice(N, sample_size, replace=False)
        for key, values in self.original_data.items():
            self.data[key] = [values[i] for i in keep_indices]
    
    def inference_downsample(self, start_idx, end_idx):
        self.logger.info("Downsampling data from index {} to {}".format(start_idx, end_idx))
        for key, values in self.original_data.items():
            self.data[key] = values[start_idx:end_idx]

    def drop(self, col='teacher_labels', value=-1, type='teacher'):
        indices = []
        data = self.teacher_data if type=='teacher' else self.student_data
        for i, array in enumerate(data[col]):
            all_neg_one = np.all(array == -1, axis=1)
            keep = np.flatnonzero(~all_neg_one).tolist()
            indices.append(keep)
        
        self.keep(indices, type=type)

    def __len__(self):
        return len(self.data['id'])

    def __getitem__(self, item):
        ret = {
            'id': self.data['id'][item],
            'input': self.data['input'][item],
            'output': self.data['output'][item] if 'output' in self.data else None,
            'input_ids': self.data['input_ids'][item],
            'output_ids': self.data['output_ids'][item] if 'output' in self.data else None,
            'align_index': self.data['align_index'][item],
            'weak_labels': self.data['weak_labels'][item],
            'sent_len': self.data['sent_len'][item],
            }
        return ret